﻿using System;
using System.Windows;

namespace Contracts
{
    /// <summary>
    /// Фабрика по производству UIControl
    /// </summary>
    public interface IUIControlFactory
    {
        /// <summary>
        /// Тип, которым может управлять контрол
        /// </summary>
        Type ControlledType { get; }
        /// <summary>
        /// Генерация нового контрола
        /// </summary>
        UIControl NewControl { get; }   
    }
    public abstract class UIControl
    {
        private object _value;
        private IModelPropertyAccessor _modelPropertyAccessor;
        private string _propertyName;
        private string _title;
        /// <summary>
        /// Отображаемое название контрола
        /// </summary>
        public virtual string Title
        {
            get
            {
                return _title;    
            }
            set
            {
                _title = value;    
            }
        }

        /// <summary>
        /// Местоположения заголовка
        /// </summary>
        public TitlePositionEnum TitlePosition=TitlePositionEnum.None;

        /// <summary>
        /// Тип, которым может управлять контрол
        /// </summary>
        public abstract Type ControlledType {get;}
        /// <summary>
        /// Элемент управления, отображаемый на форме
        /// </summary>
        public abstract FrameworkElement Control { get; }
        /// <summary>
        /// Управляемое значение
        /// </summary>
        public object Value 
        { 
            set
            {
                if (value == _value) return;
                _value = value;
                if (_modelPropertyAccessor != null)
                    _modelPropertyAccessor.SetProperty(_propertyName, value);
            }
           get
           {
               if (_modelPropertyAccessor != null)
                   _value=_modelPropertyAccessor.GetProperty(_propertyName);
               return _value;   
           }
        }

        /// <summary>
        /// Установка объекта, предоставляющего интерфейс управления свойством модели
        /// </summary>
        /// <param name="modelPropertyAccessor">Объект, предоставляющий интерфейс</param>
        /// <param name="propertyName">Имя свойства</param>
       public void SetModelPropertyAccessor(IModelPropertyAccessor modelPropertyAccessor,String propertyName)
        {
           _modelPropertyAccessor = modelPropertyAccessor;
           _propertyName = propertyName;
           _value = modelPropertyAccessor.GetProperty(propertyName);
           RefreshControl();
           
        }
        /// <summary>
        /// Указание "контролу" обновить отображаемый вид
        /// </summary>
        public abstract void RefreshControl();
    }

    /// <summary>
    /// Позиция заголовка
    /// </summary>
    public enum TitlePositionEnum
    {
        None,   //Не отображать 
        Top,    //Сверху
        Right,  //Справа
        Bottom, //Снизу
        Left    //Слева
    }
    
}
